/*
 * Copyright (C) 2005-2013 Schlichtherle IT Services.
 * All rights reserved. Use is subject to license terms.
 */
package net.java.truelicense.core;

import java.security.GeneralSecurityException;
import java.util.concurrent.Callable;
import javax.annotation.concurrent.Immutable;
import net.java.truelicense.core.auth.*;
import net.java.truelicense.core.codec.Codec;
import net.java.truelicense.core.crypto.Encryption;
import net.java.truelicense.core.io.*;

/**
 * A basic license manager.
 * <p>
 * Note that this class is immutable.
 * Unless stated otherwise, all no-argument methods need to return consistent
 * objects so that caching them is not required.
 * A returned object is considered to be consistent if it compares
 * {@linkplain Object#equals(Object) equal} or at least behaves identical to
 * any previously returned object.
 *
 * @author Christian Schlichtherle
 */
@Immutable
abstract class BasicLicenseManager
implements  LicenseVendorManager,
            LicenseConsumerManager,
            LicenseSubjectProvider,
            LicenseParametersProvider,
            StoreProvider {

    @Override
    public License create(final License input, final Sink sink)
    throws GeneralSecurityException {
        return wrap(new Callable<License>() {
            @Override public License call() throws Exception {
                authorization().clearCreate(parameters());
                return encrypt(input, sink);
            }
        });
    }

    @Override
    public License install(final Source source) throws GeneralSecurityException {
        return wrap(new Callable<License>() {
            @Override public License call() throws Exception {
                authorization().clearInstall(parameters());
                final License license = validate(source);
                Copy.copy(source, store());
                return license;
            }
        });
    }

    @Override
    public License view() throws GeneralSecurityException {
        return wrap(new Callable<License>() {
            @Override public License call() throws Exception {
                authorization().clearView(parameters());
                return decodeLicense(store());
            }
        });
    }

    @Override
    public void verify() throws GeneralSecurityException {
        wrap(new Callable<Void>() {
            @Override public Void call() throws Exception {
                authorization().clearVerify(parameters());
                validate(store());
                return null;
            }
        });
    }

    @Override
    public void uninstall() throws GeneralSecurityException {
        wrap(new Callable<Void>() {
            @Override public Void call() throws Exception {
                authorization().clearUninstall(parameters());
                final Store store = store();
                //authenticate(store);
                store.delete();
                return null;
            }
        });
    }

    //
    // Utility functions:
    //

    /**
     * Executes the given {@code task} and wraps any
     * non-{@link RuntimeException} and non-{@link GeneralSecurityException}
     * in a new {@code GeneralSecurityException}.
     *
     * @param  task the task to {@link Callable#call}.
     * @return the result of calling the task.
     * @throws RuntimeException at the discretion of the task.
     * @throws GeneralSecurityException on any other {@link Exception} thrown
     *         by the task.
     */
    private static <V> V wrap(final Callable<V> task)
    throws GeneralSecurityException {
        try { return task.call(); }
        catch (final RuntimeException ex) { throw ex; }
        catch (final GeneralSecurityException ex) { throw ex; }
        catch (final Exception ex) { throw new GeneralSecurityException(ex); }
    }

    //
    // License vendor functions:
    //

    License encrypt(final License input, final Sink sink) throws Exception {
        return compress(input, encryption().apply(sink));
    }

    License compress(final License input, final Sink sink) throws Exception {
        return encodeRepository(input, compression().apply(sink));
    }

    License encodeRepository(final License input, final Sink sink)
    throws Exception {
        final LicenseAndRepositoryProvider larp = encodeAndSign(input);
        final Repository r = larp.repository();
        codec(r.getClass()).encode(sink, r);
        return larp.license();
    }

    LicenseAndRepositoryProvider encodeAndSign(final License input)
    throws Exception {
        final License output = validate(input);
        final Repository repository = repository();
        authentication().sign(repository, output);
        return new LicenseAndRepositoryProvider() {
            @Override public License license() { return output; }
            @Override public Repository repository() { return repository; }
        };
    }

    License validate(final License input) throws Exception {
        final License output = initialize(input);
        validation().validate(output);
        return output;
    }

    License initialize(final License input) throws Exception {
        final License output = duplicate(input);
        initialization().initialize(output);
        return output;
    }

    @SuppressWarnings("unchecked")
    <V> V duplicate(final V object) throws Exception {
        final Codec codec = codec(object.getClass());
        final Store store = new MemoryStore();
        codec.encode(store, object);
        return (V) codec.decode(store);
    }

    private interface LicenseAndRepositoryProvider
    extends LicenseProvider, RepositoryProvider { }

    //
    // License consumer functions:
    //

    License validate(final Source source) throws Exception {
        final License l = decodeLicense(source);
        validation().validate(l);
        return l;
    }

    License decodeLicense(Source source) throws Exception {
        return authenticate(source).decode();
    }

    @SuppressWarnings("unchecked")
    Artifactory<License> authenticate(Source source) throws Exception {
        return (Artifactory<License>) authentication().verify(decodeRepository(source));
    }

    Repository decodeRepository(Source source) throws Exception {
        return (Repository) codec(Repository.class).decode(decompress(source));
    }

    Source decompress(Source source) {
        return compression().unapply(decrypt(source));
    }

    Source decrypt(Source source) { return encryption().unapply(source); }

    //
    // Property/factory functions:
    //

    final LicenseAuthorization authorization() {
        return parameters().authorization();
    }

    final LicenseInitialization initialization() {
        return parameters().initialization();
    }

    final LicenseValidation validation() { return parameters().validation(); }

    final Repository repository() { return parameters().repository(); }

    final Authentication authentication() {
        return parameters().authentication();
    }

    final Codec codec(Class<?> type) { return parameters().codec(type); }

    final Transformation compression() { return parameters().compression(); }

    final Encryption encryption() { return parameters().encryption(); }

    /**
     * Returns the store for the license key (optional operation).
     *
     * @throws UnsupportedOperationException If this property is called on a
     *         concrete {@link LicenseVendorManager}.
     */
    @Override public abstract Store store();
}
