/*
 * Copyright (C) 2005-2013 Schlichtherle IT Services.
 * All rights reserved. Use is subject to license terms.
 */
package net.java.truelicense.core.auth;

import java.nio.charset.Charset;
import java.security.*;
import java.util.Arrays;
import javax.annotation.*;
import net.java.truelicense.core.codec.*;
import net.java.truelicense.core.io.*;
import net.java.truelicense.obfuscate.Obfuscate;
import org.apache.commons.codec.binary.Base64;

/**
 * A generic repository.
 *
 * @author Christian Schlichtherle
 */
public class GenericRepository implements Repository {

    @Obfuscate
    private static final String BASE64 = "base64";

    @Obfuscate
    private static final String SIGNATURE_TYPE = "application/octet-stream";

    @Obfuscate
    private static final String SIGNATURE_TRANSFER_ENCODING = BASE64;

    private static final Base64 base64 = new Base64();

    private final Codec codec;

    private @CheckForNull byte[] artifactData, signatureData;

    /**
     * The character set as specified by - or implied from - the codec.
     * If this is {@code null}, then the codec uses an unknown character set or
     * does not produce text output and needs a Base64 transformation.
     */
    private final @CheckForNull Charset charset;

    /**
     * Constructs a generic repository.
     *
     * @param codec the codec for the artifact.
     */
    public GenericRepository(final Codec codec) {
        this.charset = Codecs.charset(this.codec = codec);
    }

    /** Returns the codec which was provided to the constructor. */
    public Codec getCodec() { return codec; }

    public String getArtifactType() { return getCodec().contentType(); }

    public String getArtifactTransferEncoding() {
        return null == charset ? BASE64 : getCodec().contentTransferEncoding();
    }

    public String getArtifactBody() {
        final byte[] artifactData = this.artifactData;
        return null == artifactData
                ? null
                : null == charset
                    ? base64.encodeToString(artifactData)
                    : new String(artifactData, charset);
    }

    public void setArtifactBody(final String body) {
        artifactData = null == body
                ? null
                : null == charset
                    ? base64.decode(body)
                    : body.getBytes(charset);
    }

    public String getSignatureType() { return SIGNATURE_TYPE; }

    public String getSignatureTransferEncoding() {
        return SIGNATURE_TRANSFER_ENCODING;
    }

    public String getSignatureBody() {
        return base64.encodeToString(signatureData);
    }

    public void setSignatureBody(final String body) {
        signatureData = base64.decode(body);
    }

    @Override
    public <A> Artifactory<A> sign(
            final Signature engine,
            final PrivateKey key,
            final @CheckForNull A artifact)
    throws Exception {
        final MemoryStore store = new MemoryStore();
        final Codec codec = getCodec();
        codec.encode(store, artifact);
        final byte[] artifactData = store.data();
        engine.initSign(key);
        engine.update(artifactData);
        final byte[] signatureData = engine.sign();
        this.artifactData = artifactData;
        this.signatureData = signatureData;
        return new EncodedArtifact<A>(codec, store);
    }

    /**
     * {@inheritDoc}
     *
     * @throws IllegalStateException if the properties do not match the
     *         properties used when signing.
     */
    @Override
    public Artifactory<?> verify(final Signature engine, final PublicKey key)
    throws Exception {
        final Codec codec = getCodec();
        engine.initVerify(key);
        final byte[] artifactData = this.artifactData;
        engine.update(artifactData);
        if (!engine.verify(signatureData))
            throw new RepositoryIntegrityException();
        return new EncodedArtifact<Object>(codec, artifactData);
    }

    @Override @SuppressWarnings("AccessingNonPublicFieldOfAnotherObject")
    public boolean equals(final Object obj) {
        if (this == obj) return true;
        if (!(obj instanceof GenericRepository)) return false;
        final GenericRepository that = (GenericRepository) obj;
        return  Arrays.equals(this.artifactData, that.artifactData) &&
                Arrays.equals(this.signatureData, that.signatureData);
    }

    @Override public int hashCode() {
        int c = 17;
        c = 31 * c + Arrays.hashCode(artifactData);
        c = 31 * c + Arrays.hashCode(signatureData);
        return c;
    }
}
